/*
Clock Data Recoverary Control Training IP in CDR mode 4
*/

module CORECDR4CNTL_TIP (
                         input          cdr_start,  // active low reset in sclk1 domain, generated from delay code TIP
                         input          sclk0,  // [0] for RX, divclk from lanectrl
                         input          stream_start,  // initial flag ready or stream start from flag generator TIP or SGMII IP
                         input          early_n,  // active high early flag from IOGN data eye monitor or flag generator TIP
                         input          late_n,  // active high late flag from IOGN data eye monitor or flag generator TIP
                         input          dll_valid_code,  // DLL delay code valid indicator from delay code TIP for clean clock domain transfer
                         input [6:0]    dll_delay_code,  // DLL delay code for 90 degree from delay code TIP
                         output         cdr_ready,  // ready indicator to SGMII IP
                         output reg     clr_flags_n,  // active high reset to IOGN data eye monitor
                         output [2:0]   eyewidth_in_lane,  // parameterized data eye window width
                         output reg [10:0]  sela_lane,  // clka phase/delay setting
                         output reg [10:0]  selb_lane,  // clkb phase/delay setting
                         output reg         switch_lane);  // glitchless switch cdrclk to nxtclk

parameter       ST_RSWT = 2'b00;  // right shift switch
parameter       ST_RWAT = 2'b01;  // right wait for flag
parameter       ST_LSWT = 2'b10;  // left shift switch
parameter       ST_LWAT = 2'b11;  // left wait for flag

parameter       STEP = 3;  // jumping step
//parameter     EYEWIDTH = 3'b111;  // +/-7 eyewidth
parameter       EYEWIDTH = 3'b001;  // +/-1 eyewidth

reg [1:0]       start_cnt;
reg [1:0]       rst_n;  // sync-off internal reset on sclk0 domain
reg [1:0]       coarse_sel;  // coarse phase selection
reg [7:0]       fine_sel;  // fine phase selection
reg [1:0]       tune_st;  // state_machine
reg [3:0]       cnt;
reg [1:0]       early_flag;  // synchronizer
reg [1:0]       late_flag;  // synchronizer
reg [1:0]       valid_flag;  // synchronizer
reg [6:0]       dll_90_code;  // 90 degree code from DLL synchronizer
reg             clr_flag;
reg [3:0]       cdr_ready_reg;  // wait states for cdr_ready

assign cdr_ready = cdr_ready_reg[0];  // wait several cycles after first switch so that RX data can be pipe-cleaned
assign eyewidth_in_lane = EYEWIDTH;

always @(posedge sclk0 or negedge cdr_start) begin  // RX clock
   if (!cdr_start) rst_n <= 0;  // async-on
   else if (stream_start) rst_n <= {1'b1, rst_n[1]};  // sync-off after the first flags are generated
end

always @(posedge sclk0 or negedge rst_n[0] or posedge clr_flag) begin  // RX clock
   if (!rst_n[0]) begin
      early_flag <= 0;
      late_flag <= 0;
   end
   else if (clr_flag) begin
      early_flag <= 0;
      late_flag <= 0;
   end
   else begin
      early_flag <= {early_n, early_flag[1]};
      late_flag <= {late_n, late_flag[1]};
   end
end

always @(posedge sclk0 or negedge rst_n[0]) begin  // RX clock
   if (!rst_n[0]) begin
      cdr_ready_reg <= 0;
      valid_flag <= 2'b11;
      dll_90_code <= 0;
   end
   else begin
      if (sela_lane[10]) cdr_ready_reg[3] <= 1'b1;  // after the first switch, the CDR is ready
      cdr_ready_reg[2:0] <= cdr_ready_reg[3:1];

      valid_flag <= {dll_valid_code, valid_flag[1]};  // sync
      if (valid_flag[0]) dll_90_code <= dll_delay_code;  // when valid, the dll_delay_code is guaranteed stable
   end
end

always @(posedge sclk0 or negedge rst_n[0]) begin  // RX clock
   if (!rst_n[0]) begin
      clr_flags_n <= 1'b1;  // initial clear all early/late flags
      sela_lane <= 0;
      selb_lane <= 0;
   end
   else begin
      clr_flags_n <= (clr_flag | switch_lane);
      if (clr_flag) begin
         if (sela_lane[10]) sela_lane[9:0] <= {coarse_sel[1:0], fine_sel[7:0]};
         if (!selb_lane[10]) selb_lane[9:0] <= {coarse_sel[1:0], fine_sel[7:0]};
      end
      else if (cdr_ready_reg[2:0] == 3'b110) sela_lane[9:0] <= {coarse_sel[1:0], fine_sel[7:0]};  // initial condition
      
      if (switch_lane) begin
         sela_lane[10] <= ~sela_lane[10];  // cdrclk_selb
         selb_lane[10] <= ~selb_lane[10];  // nxtclk_sela
      end
   end
end

always @(posedge sclk0 or negedge rst_n[0]) begin  // RX clock
   if (!rst_n[0]) begin
      coarse_sel <= 0;
      fine_sel <= 0;
      tune_st <= ST_RSWT;
      cnt <= 0;
      clr_flag <= 0;
      switch_lane <= 0;
   end
   else case (tune_st)
      ST_RSWT : if (!cdr_ready_reg[3]) begin  // initial condition
                   if (early_flag[0] || late_flag[0]) begin  // initial noisy
                      if (early_flag[0]) coarse_sel <= coarse_sel - 1;  // 90 degree ahead
                      fine_sel <= dll_90_code[6:1];  // 45 degree left or right shift
                      tune_st <= ST_LSWT;
                      clr_flag <= 1'b1;
                      cnt <= 0;
                   end
                   else if (& cnt) begin  // long wait for flags in initial search
                      cnt <= 2;
                      clr_flag <= 1'b1;
                   end
                   else begin
                      cnt <= cnt + 1;
                      clr_flag <= 1'b0;  // one cycle assertion
                   end
                   
                   if (switch_lane) switch_lane <= 1'b0;
                   else if ((& cnt) && !(early_flag[0] || late_flag[0])) switch_lane <= 1'b1;
                end
                else begin
                   if (cnt == 4) begin
                      tune_st <= ST_RWAT;
                      cnt <= 0;
                      clr_flag <= 1'b1;
                   end
                   else begin
                      cnt <= cnt + 1;
                      clr_flag <= 1'b0;
                   end
                   if (switch_lane) switch_lane <= 1'b0;
                   else if (cnt == 2) switch_lane <= 1'b1;
                end
      ST_LSWT : begin
                   if (cnt == 4) begin
                      tune_st <= ST_LWAT;
                      cnt <= 0;
                      clr_flag <= 1'b1;
                   end
                   else begin
                      cnt <= cnt + 1;
                      clr_flag <= 1'b0;
                   end
                   if (switch_lane) switch_lane <= 1'b0;
                   else if (cnt == 2) switch_lane <= 1'b1;
                end
      ST_RWAT : if (late_flag[0]) begin  // confirm move right after right move
                   if ((fine_sel + STEP) <= dll_90_code) begin  // one full step later
                      fine_sel <= fine_sel + STEP; 
                   end
                   else begin
                      coarse_sel <= coarse_sel + 1;  // 90 degree later
                      fine_sel <= fine_sel + STEP - dll_90_code; 
                   end
                   tune_st <= ST_RSWT;
                   clr_flag <= 1'b1;
                end
                else if (early_flag[0]) begin  // confirm move left after right move
                   if (fine_sel >= (STEP - 1)) begin  // one tap less back
                      fine_sel <= fine_sel - (STEP - 1); 
                   end
                   else begin
                      coarse_sel <= coarse_sel - 1;  // 90 degree ahead
                      fine_sel <= dll_90_code - (STEP - 1) + fine_sel; 
                   end
                   tune_st <= ST_LSWT;
                   clr_flag <= 1'b1;
                end
                else begin
                   clr_flag <= 1'b0;
                end
      ST_LWAT : if (early_flag[0]) begin  // confirm move left after left move
                   if (fine_sel >= STEP) begin  // one full step ahead
                      fine_sel <= fine_sel - STEP; 
                   end
                   else begin
                      coarse_sel <= coarse_sel - 1;  // 90 degree ahead
                      fine_sel <= dll_90_code - STEP + fine_sel; 
                   end
                   tune_st <= ST_LSWT;
                   clr_flag <= 1'b1;
                end
                else if (late_flag[0]) begin  // confirm move right after left move
                   if ((fine_sel + STEP - 1) <= dll_90_code) begin  // one tap less later
                      fine_sel <= fine_sel + STEP - 1; 
                   end
                   else begin
                      coarse_sel <= coarse_sel + 1;  // 90 degree later
                      fine_sel <= fine_sel + STEP - 1 - dll_90_code; 
                   end
                   tune_st <= ST_RSWT;
                   clr_flag <= 1'b1;
                end
                else begin
                   clr_flag <= 1'b0;
                end
   endcase
end

endmodule